////////////////////////////
////////// base ///////////
///////////////////////////

nodes = null;
links = null;
feature_vector = null;
con_matrix = null;
sankey = null;
fingerprint_view = null;
Filter = null;
play = null;
infobar = null;
info_view = null;

function make_data(data) {
    info_view = new InfoView(data.info, d3.select('.info_view').append("table"));
    feature_vector = new FeatureVector(data.feature_vector);
    nodes = new NodeArray(data.nodes, feature_vector);
    links = new LinkArray(nodes);

    con_matrix = new Matrix(nodes.size(), nodes.size(), -1);
    links.getData().forEach(function (l, i) {
        con_matrix.set(l.from.id, l.to.id, i);
    });
}

function make_diagrams() {
    if (infobar == null) infobar = new Infobar();

    if (fingerprint_view == null) fingerprint_view = new FingerprintView(d3.select('.fingerprints_view').append("svg"),
        d3.select('.fingerprints_view_common').append("svg"));

    if (sankey == null) sankey = new SankeyDiagram(nodes, links, feature_vector, con_matrix, fingerprint_view, infobar);//,Fingerprints);
    else sankey.reset(nodes, links, feature_vector, con_matrix, fingerprint_view)
    if (Filter == null) Filter = new FilterSelection(nodes.size(), feature_vector, sankey, d3.select('.filter_fingerprint').append("svg"));
}

function do_page() {
    d3.json(FILENAME, function (data) {
        make_data(data);
        make_diagrams()
    });

    if (document.getElementById('checkbox_reload').checked) var timeout = setTimeout(function () {
        do_page();
    }, 1000);
}


/////////////////////////////////////
////////// feature vector ///////////
/////////////////////////////////////

class FeatureVector extends Array {

    constructor(data) {
        super();
        data.forEach(featureData => this.push({
            type: featureData.featureType,
            serial: featureData.ligFeatureSerial + ', ' + featureData.protFeatureSerial
        }))
    }

    getLength() {
        return this.length;
    }
}

///////////////////////////////////////
////////// Filter Selection ///////////
///////////////////////////////////////

class FilterSelection {

    constructor(nodes_count, feature_vector, SankeyDiagram, svg) {
        this.SankeyDiagram = SankeyDiagram;
        this.feature_vector = feature_vector;
        this.svg = new FingerprintSVG(svg, true, false);
        this.filter_slider = document.getElementById("filter_slider");
        this.slider_value = document.getElementById("slider_value");

        const self = this;
        this.filter_slider.max = nodes_count;

        this.filter_slider.oninput = function () {
            self.slider_value.innerHTML = this.value;
            self.SankeyDiagram.set_filtert_node(this.value)
            self.node.getFingerprintType().clearBits();
            self.svg.update();
        }

        this.node = new Node({}, this.feature_vector);
        this.node.getFingerprintType().clearBits();
        this.svg.addNode(this.node);
        const bits = this.svg.getBitsRect();
        bits.enter().selectAll("rect").on("click", function (d, i) {
            self.on_click(d, i, d3.event.shiftKey);
        });

    }

    on_click(d, i) {
        const self = this;
        this.node.getFingerprintType().toggleFingerprintBit(i);
        this.svg.update();
        const bits = this.svg.getBitsRect();
        bits.enter().selectAll("rect").on("click", function (d, i) {
            self.on_click(d, i, d3.event.shiftKey);
        });
        this.SankeyDiagram.filter_by_fingerprint(this.node.getFingerprint())
    }
}

//////////////////////////////////
////////// fingerprint ///////////
//////////////////////////////////

featureColor = {
    H: 'yellow',
    AR: 'blue',
    HBA: 'red',
    HBD: 'green',
    PI: 'cyan',
    NI: 'pink',
    XBD: 'grey',
    none: 'white'
};

class Fingerprint {

    constructor(feature_vector, fingerprint = "") {
        this.feature_vector = feature_vector;
        this.fingerprint = fingerprint;
    }

    static common_fingerprint(fingerprint1, fingerprint2) {
        let ret = '';
        const l = fingerprint1.length;

        for (let i = 0; i < l; i++) {
            if (fingerprint1[i] === '1' && fingerprint2[i] === '1') ret += '1';
            else ret += '0';
        }
        return ret;
    }

    static is_zero(fingerprint) {
        const l = fingerprint.length;
        for (let i = 0; i < l; i++) {
            if (fingerprint[i] === '1') return false;
        }
        return true;
    }

    static a_has_all_b(a, b) {
        let ret = true;
        for (let i = 0; i < a.length; i++) {
            if (b[i] === '1' && a[i] === '0') ret = false;
        }
        return ret;
    }

    getLength() {
        return this.fingerprint.length;
    }

    getFingerprint() {
        return this.fingerprint;
    }

    getFingerprintToColor(sorted = false) {
        const self = this;
        const ret = [];
        if (!sorted) {
            this.feature_vector.forEach(function (fv, i) {
                ret.push({
                    type: fv.type,
                    serial: fv.serial,
                    color: featureColor[fv.type],
                    on: self.fingerprint[i] === '1'
                });

            });
        } else {
            ret.push(this.getColorFor('H'));
            ret.push(this.getColorFor('AR'));
            ret.push(this.getColorFor('HBA'));
            ret.push(this.getColorFor('HBD'));
            ret.push(this.getColorFor('PI'));
            ret.push(this.getColorFor('NI'));
            ret.push(this.getColorFor('XBD'));
        }
        return ret;
    }

    getColorFor(type) {
        let on = 0;
        let off = 0;
        for (let i = 0; i < this.feature_vector.getLength(); i++) {
            if (this.feature_vector[i].type === type) {
                if (+this.fingerprint[i] === 1) on++;
                else off++;
            }
        }
        return {color: featureColor[type], on: on, off: off};
    }

    getLabels() {
        const ret = [];

        this.feature_vector.forEach(function (fv) {
            ret.push(fv.type);

        });

        //remove doubles
        for (let i = ret.length - 1; i >= 0; i--) {
            if (ret[i] === ret[i - 1]) ret[i] = "";
        }
        return ret;
    }

    clearBits() {
        this.fingerprint = '0';
        this.fingerprint = this.fingerprint.repeat(this.feature_vector.getLength());
    }

    setBits() {
        this.fingerprint = '1';
        this.fingerprint = this.fingerprint.repeat(this.feature_vector.getLength());
    }

    toggleFingerprintBit(nr) {
        let fp = this.fingerprint;
        if (fp[nr] === '1') fp = fp.substr(0, nr) + '0' + fp.substr(nr + 1);
        else fp = fp.substr(0, nr) + '1' + fp.substr(nr + 1);
        this.fingerprint = fp;
    }
}

//////////////////////////////////////
////////// fingerprint svg ///////////
//////////////////////////////////////

const FINGERPRINTS_MARGIN_LEFT = 50;
const FINGERPRINTS_MARGIN_TOP = 20;
const FINGERPRINTS_MARGIN_BOTTOM = 5;

const FINGERPRINT_BITSIZE = 20;


class FingerprintSVG {
    constructor(svg, showXLabel = true, showYLabel = true) {
        this.svg_base = svg;
        this.showYLabel = showYLabel;
        this.showXLabel = showXLabel;

        this.data = [];// fingerprint

        this.svg_base
            .attr("class", "fingerprints_view")
            .attr("width", 1750 + 1500)//this.fingerprint.width())
            .attr("height", 0)//this.fingerprint.height())
        this.svg = this.svg_base.append("g")
            .attr("transform", "translate(" + FINGERPRINTS_MARGIN_LEFT + "," + FINGERPRINTS_MARGIN_TOP + ")");

        this.xlabel_data = [];
        this.ylabel_data = []

        this.tip = d3.tip()
            .attr('class', 'tip_fingerprint')
            .style("visibility", "visible")
            .offset([FINGERPRINT_BITSIZE * 2, 0])
            .html(function (d) {
                return d.type + ", " + d.serial;
            });
        this.tip(this.svg.append("g"));

        this.bits = null;
    }

    data_to_array(data) {
        const ret = [];
        data.forEach(function (d, y) {
            const col = d.getFingerprintType().getFingerprintToColor();
            col.forEach(function (c, x) {
                c.x = x;				// extend object
                c.y = y;
                ret.push(c)
            });

        });
        return ret;
    }

    makeXlabels() {
        return this.data[0].getFingerprintType().getLabels();
    }

    makeYlabels(data) {
        const ret = [];
        const x = this.fingerprint_svg_width() + 10;
        data.forEach(function (d, i) {
            ret.push({
                text: (d.node_frames.length < 50 ? "Node " + d.id + ", Frame(s) " + d.node_frames : "Node " + d.id + ", Frame(s) " + d.node_frames.slice(0, 20) + '...'),
//                    text: "Node " + d.id + ", Frame(s) " + d.node_frames,
                    x: x,
                    y: i * FINGERPRINT_BITSIZE + FINGERPRINTS_MARGIN_TOP - 2
                }
            );
        });
        return ret;
    }

    fingerprint_svg_width() {
        return this.data[0].getFingerprintType().getLength() * FINGERPRINT_BITSIZE;
    }

    update(only_bits = false) {
        const data = this.data_to_array(this.data);
        const self = this;

        if (only_bits === false) {

            if (data.length > 0) this.ylabel_data = this.makeYlabels(this.data);
            if (this.xlabel_data.length === 0 && data.length > 0 || data.length === 0) {
                if (data.length === 0) {
                    this.xlabel_data = []
                    this.ylabel_data = [];
                } else {
                    this.xlabel_data = this.makeXlabels();
                }
            }

            if (this.showXLabel) {
                const xLabels = this.svg.selectAll(".fingerprint_view_bit_label").data(this.xlabel_data);
                xLabels.data(this.xlabel_data)
                    .enter().append("text")
                    .text(d => d)
                    .attr("x", function (d, i) {
                        return i * FINGERPRINT_BITSIZE;
                    })
                    .attr("y", 0)
                    .attr("class", "fingerprint_view_bit_label")
                    .attr("transform", "translate(" + FINGERPRINT_BITSIZE / 5 + ", -6)")
                xLabels.exit().remove();
            }

            if (this.showYLabel) {
                const yLabels = this.svg.selectAll(".fingerprint_view_bit_Ylabel").data(this.ylabel_data, function (d) {
                    return d.text + ':' + d.x + ':' + d.y;
                });
                yLabels.enter().append("text")
                    .text(d => d.text)
                    .attr("x", d => d.x)
                    .attr("y", d => d.y)
                    .attr("class", "fingerprint_view_bit_Ylabel")
                yLabels.exit().remove();
            }
        }

        this.bits = this.svg.selectAll(".fingerprint_view_bit").data(data, function (d) {
            return d.x + ':' + d.y + ':' + d.on;
        });
        this.bits.enter().append("rect")
            .attr("x", d => d.x * FINGERPRINT_BITSIZE)
            .attr("y", d => d.y * FINGERPRINT_BITSIZE)
            .attr("rx", 5)
            .attr("ry", 5)
            .attr("class", "fingerprint_view_bit")
            .classed("active", d => +d.on === 1)
            .attr("width", FINGERPRINT_BITSIZE)
            .attr("height", FINGERPRINT_BITSIZE)
            .style("fill", d => d.color)
            .on('mouseover', self.tip.show)
            .on('mouseout', self.tip.hide)
        this.bits.exit().remove();

        this.fit_svg();
    }

    fit_svg() {
        this.svg_base.attr("height", this.data.length * FINGERPRINT_BITSIZE + FINGERPRINTS_MARGIN_TOP + FINGERPRINTS_MARGIN_BOTTOM);
    }

    addNode(d) {
        this.data.push(d);
        this.update();
    }

    removeNode(d) {
        const pos = this.data.indexOf(d);
        this.data.splice(pos, 1);
        this.update();
    }

    removeAll() {
        this.data = [];
        this.update();
    }

    getBitsRect() {
        return this.bits;
    }

    setBit(d, i, v) {
        console.log(this.data[0].setFingerprintBit(i, v))
        this.update();
    }
}

///////////////////////////////////////
////////// fingerprint view ///////////
///////////////////////////////////////

class FingerprintView extends FingerprintSVG {

    constructor(svg, svg_common) {
        super(svg, false, true);
        this.svg_base_common = new FingerprintSVG(svg_common, false, true);
    }

    setCommonNode(d) {
        this.svg_base_common.removeAll();
        this.svg_base_common.addNode(d);
    }

    removeCommonNode() {
        this.svg_base_common.removeAll();
    }

    removeAll() {
        super.removeAll();
        this.svg_base_common.removeAll();
    }

}

///////////////////////////////
////////// info bar ///////////
///////////////////////////////

class Infobar {

    constructor() {
        this.topMargin = 0;
        this.infobar = document.getElementById('infobar');
        this.infolist = [];
    }

    showinfoText(type, text) {
        if (type === "error") this.infobar.classList.add('error');
        else {
            this.infobar.classList.remove('error');
            if (this.infolist.includes(text)) return;
            this.infolist.push(text);
        }
        this.infobar.innerHTML = text;
        this.showinfo();
    }

    showinfo() {
        const self = this;
        document.body.style.marginTop = (this.topMargin++) + 'px';
        this.infobar.style.top = (this.topMargin - 30) + 'px';
        if (this.topMargin < 30) setTimeout(function () {
            self.showinfo();
        }, 10);
    }

    hideinfo() {
        const self = this;
        document.body.style.marginTop = (--this.topMargin) + 'px';
        this.infobar.style.top = (this.topMargin - 30) + 'px';
        if (this.topMargin > 0) setTimeout(function () {
            self.hideinfo();
        }, 10);
    }

}

/////////////////////////////////////
////////// info view ///////////
/////////////////////////////////////

class InfoView {

    constructor(data, table) {

        console.log(data);
        console.log(JSON.stringify(data.proportion));

        const columns = ['Variance of the projection (%)'];
        const thead = table.append('thead');
        const tbody = table.append('tbody');

        // append the header row
        thead.append('tr')
            .selectAll('th')
            .data(columns).enter()
            .append('th')
            .text(function (column) {
                return column;
            });

        // create a row for each object in the data
        const rows = tbody.selectAll('tr')
            .data(Object.keys(data))
            .enter()
            .append('tr');

        // create a cell in each row for each column
        const cells = rows.selectAll('td')
            .data(function (row) {
                return columns.map(function (column) {
                    return {column: column, value: row[column]};
                });
            })
            .enter()
            .append('td')
            .text(function (d) {
                return (parseFloat(data.proportion) * 100).toFixed(1);
            });

    }
}

///////////////////////////
////////// link ///////////
///////////////////////////

class LinkArray {

    constructor(in_data) {
        const self = this;
        this.data = [];

        in_data.getData().forEach(function (n1) {
            const cid = n1.getChildIds();
            cid.forEach(function (c) {
                const n2 = in_data.getWithId(c);
                self.data.push(new Link(n1, n2));
            });
        });

    }

    getData() {
        return this.data;
    }

    size() {
        return this.data.length;
    }
}

class Link {

    constructor(node1, node2) {
        this.from = {x: node1.getX(), y: node1.getY(), id: node1.getId()}
        this.to = {x: node2.getX(), y: node2.getY(), id: node2.getId()}
        this.active = false;
    }

}

//////////////////////////////
////////// matrix ////////////
//////////////////////////////

class Matrix extends Array {

    constructor(rows, cols, initVal = 0) {
        super();
        this.rows = rows;
        this.cols = cols;
        for (let i = 0; i < rows; i++) {
            const a = [];
            for (let j = 0; j < cols; j++) a.push(initVal);
            this.push(a);
        }
    }

    set(r, c, v) {
        this[r][c] = v;
    }

    rows() {
        return this.rows;
    }

    cols() {
        return this.cols;
    }

    get(r, c) {
        return this[r][c];
    }
}

///////////////////////////
////////// node ///////////
///////////////////////////

class NodeArray {

    constructor(data, feature_vector) {
        this.data = [];
        this.feature_vector = feature_vector;
        data.forEach(d => this.data.push(new Node(d, feature_vector)));
    }

    size() {
        return this.data.length;
    }

    max_x() {
        return this.data.reduce((max, p) => p.x > max ? p.x : max, this.data[0].x);
    }

    max_y() {
        return this.data.reduce((max, p) => p.y > max ? p.y : max, this.data[0].y);
    }

    getMinX() {
        return this.data.reduce((min, p) => p.x < min ? p.x : min, this.data[0].x);
    }

    getMaxX() {
        return this.data.reduce((max, p) => p.x > max ? p.x : max, this.data[0].x);
    }

    getMinY() {
        return this.data.reduce((min, p) => p.y < min ? p.y : min, this.data[0].y);
    }

    getMaxY() {
        return this.data.reduce((max, p) => p.y > max ? p.y : max, this.data[0].y);
    }

    getMinSize() {
        return this.data.reduce((min, p) => p.size < min ? p.size : min, this.data[0].size);
    }

    getMaxSize() {
        return this.data.reduce((max, p) => p.size > max ? p.size : max, this.data[0].size);
    }

    getMinFillColor() {
        return this.data.reduce((min, p) => p.fillColor < min ? p.fillColor : min, this.data[0].fillColor);
    }

    getMaxFillColor() {
        return this.data.reduce((max, p) => p.fillColor > max ? p.fillColor : max, this.data[0].fillColor);
    }

    max_frame() {
        return this.data.reduce((max, p) => p.max_frame() > max ? p.max_frame() : max, this.data[0].max_frame());
    }

    getWithId(id) {
        return this.data.filter(d => {
            return d.getId() === id;
        })[0];
    }

    getWithFingerprint(fingerprint) {
        return this.data.filter(d => {
            return d.getFingerprint() === fingerprint;
        })[0];
    }

    getData() {
        return this.data;
    }
}

class Node {

    constructor(data, feature_vector) {

        if (data.size > 0 && data.node_frames_length > 0) {
            this.node_frames = data.node_frames;
        } else {
            this.node_frames = [];
        }
        this.id = data.node_id;
        this.fingerprint = new Fingerprint(feature_vector, data.node_fingerprint);
        this.x = data.x //* NODE_DISTANCE_X;
        this.y = data.y     //* NODE_DISTANCE_Y;
        this.size = data.size;
        this.child_id = data.child_id;
        this.fillColor = data.fillColor;
        this.feature_vector = feature_vector;
        this.active = false;
        this.filtered = false;
        this.level = this.x;
    }

    getFingerprintType() {
        return this.fingerprint;
    }

    getFingerprint() {
        return this.fingerprint.getFingerprint();
    }

    has_frame(frame) {
        if (this.node_frames.length === 0) return false;
        if (Array.isArray(this.node_frames))
            return this.node_frames.includes(frame);
        else
            return this.node_frames === frame;
    }

    max_frame() {
        return this.node_frames.reduce((max, p) => p > max ? p : max, this.node_frames[0]);
    }

    getId() {
        return this.id;
    }

    setId(id) {
        this.id = id;
    }

    getX() {
        return this.x;
    }

    getY() {
        return this.y;
    }

    getChildIds() {
        return this.child_id;
    }

    getLevel() {
        return this.level;
    }

    getActive() {
        return this.active;
    }

    setActive(active) {
        this.active = active;
    }

    getFiltered() {
        return this.filtered;
    }

    getSize() {
        return this.size;
    }
}

/////////////////////////////////////
////////// sankey diagram ///////////
/////////////////////////////////////

const SankeyDiagram_margin = {top: 20, right: 5, bottom: 10, left: 0};

class SankeyDiagram {

    constructor(nodes, links, feature_vector, connection_link_ids, fingerprints_view, infobar) {
        this.tooltip = new NodeTooltip(feature_vector);
        this.infobar = infobar;
        this.reset(nodes, links, feature_vector, connection_link_ids, fingerprints_view);
    }

    reset(nodes, links, feature_vector, connection_link_ids, fingerprints_view) {
        this.nodes = nodes;
        this.nodes_data = nodes.getData();
        this.links = links;
        this.links_data = links.getData();

        if (this.svg == null) {
            this.make_svg_main_tree();
        }

        this.connection_link_ids = connection_link_ids;
        this.fingerprints_view = fingerprints_view;
        this.clicked_nodes = [];
        this.update_tree();
    }

    make_svg_main_tree() {
        const self = this;
        const w = window.innerWidth;
        const h = window.innerHeight;

        d3.select('.tree').on("click", function (d, i) {
            if (d3.event.target.tagName === "svg") self.on_node_click(d, i, d3.event.shiftKey);
        })

        this.svg = d3.select('.tree').append('svg')
            .attr("width", this.width())
            .attr("height", this.height())
            .call(d3.zoom().on("zoom", function () {
                self.svg.attr("transform", d3.event.transform)
            }))
            .append("g")
            .attr("transform", "translate(" + SankeyDiagram_margin.left + "," + SankeyDiagram_margin.top + ")")

        this.svg.append("g").attr("class", "link")
        this.svg.append("g").attr("class", "node")

        this.svg.append("g").classed("chartMain", true).call(this.tooltip.get_tooltip());

        var link = this.svg.select(".link").selectAll("path").data(this.links_data)
        link.enter().append("path")
            .attr("stroke-width", "0.2px")
            .attr("d", function () {						//this.path);
                return "M" + w / 2 + "," + h / 2
                    + "C" + w / 2 + "," + h / 2
                    + " " + w / 2 + "," + h / 2
                    + " " + w / 2 + "," + h / 2;
            })

        const circle = this.svg.select(".node").selectAll("circle").data(this.nodes_data);//, function(d) { return d; })
        circle.enter().append("circle")
            .attr("cx", w / 2)
            .attr("cy", h / 2)
            .attr("r", 1)
            .attr("fill", 'black')
            .attr("stroke-width", "0px")
            .on('mouseover', function (d) {
                self.draw_tooltip(d);
            })
            .on("mousemove", function () {
                self.move_tooltip();
            })
            .on('mouseout', function () {
                self.hide_tooltip();
            })
            .on("click", function (d, i) {
                self.on_node_click(d, i, d3.event.shiftKey);
            });

    }

    update_tree() {
        const yScale = d3.scaleLinear().domain([this.nodes.getMinY(), this.nodes.getMaxY()]).range([0, this.height()]);
        const xScale = d3.scaleLinear().domain([this.nodes.getMinX(), this.nodes.getMaxX()]).range([0, this.width()]);
        let colorScaleFill;
        let sizeScale;
        const self = this;

        function path(d) {
            const curvature = .5;
            const x0 = d.from.x,//+node_data.r,
                x1 = d.to.x,//-node_data.r,
                xi = d3.interpolateNumber(x0, x1),
                x2 = xi(curvature),
                x3 = xi(1 - curvature),
                y0 = d.from.y,
                y1 = d.to.y;
            return "M" + xScale(x0) + "," + yScale(y0)
                + "C" + xScale(x2) + "," + yScale(y0)
                + " " + xScale(x3) + "," + yScale(y1)
                + " " + xScale(x1) + "," + yScale(y1);
        }

        colorScaleFill = d3.scaleLinear().domain([0, this.nodes.getMaxFillColor()]).range(d3["schemeCategory10"])
        const link = this.svg.select(".link").selectAll("path").data(this.links_data);
        link.exit().remove();
        link.classed("active", d => d.active).classed("filtered", d => d.filtered)
        link.transition().duration(500).attr("d", function (d) {
            return path(d);
        });
        link.enter().append("path")
            .transition().duration(500).attr("d", function (d) {
            return path(d);
        });

        const circle = this.svg.select(".node").selectAll("circle").data(this.nodes_data);//, function(d) { return d; })
        circle.exit().remove();
        circle.classed("active", d => d.active).classed("filtered", d => d.filtered)
        circle.transition().duration(500)
            .attr("cx", d => xScale(d.x))
            .attr("cy", d => yScale(d.y))
            .attr("r", d => yScale(d.getSize()))
            .attr("fill", d => colorScaleFill(d.fillColor))
        circle.enter().append("circle")
            .transition().duration(500)
            .attr("cx", d => xScale(d.x))
            .attr("cy", d => yScale(d.y))
            .attr("r", d => yScale(d.getSize()))
            .on('mouseover', function (d) {
                self.draw_tooltip(d);
            })
            .on("mousemove", function () {
                self.move_tooltip();
            })
            .on('mouseout', function () {
                self.hide_tooltip();
            })
            .on("click", function (d, i) {
                self.on_node_click(d, i, d3.event.shiftKey);
            });
    }

    width() {
        return window.innerWidth - SankeyDiagram_margin.left - SankeyDiagram_margin.right;
    }

    height() {
        return window.innerHeight - SankeyDiagram_margin.top - SankeyDiagram_margin.bottom;
    }

    draw_tooltip(d) {
        this.tooltip.draw(d);
    }

    move_tooltip() {
        this.tooltip.move();
    }

    hide_tooltip() {
        this.tooltip.hide();
    }

    deselect_nodes() {
        this.nodes_data.forEach(d => d.active = false);
        d3.selectAll(".node").selectAll("circle").classed("active", d => d.active).data(this.nodes_data);
    }

    deselect_links() {
        this.links_data.forEach(d => d.active = false);
        d3.selectAll(".link").selectAll("path").classed("active", d => d.active).data(this.links_data);
    }

    link_between_nodes(a, b) {
        for (let i = 0; i < this.links_data.length; i++) {
            if (this.links_data[i].from.id === a && this.links_data[i].to.id === b) return i;
            if (this.links_data[i].from.id === b && this.links_data[i].to.id === a) return i;
        }
        this.infobar.showinfoText("error", "no link between nodes " + a + ":" + b)
        console.log("ERROR:link_between_nodes Line:" + new Error().lineNumber)
        return links[0];
    }

    get_from_to(from, to) {

        let i;
        const connections = [];
        const open_wege = [];
        const closed_wege = [];

        for (i in this.nodes_data) {
            const id = this.nodes_data[i].getId();
            connections[id] = this.connected_nodes_on_node(id, "source");
        }

        open_wege.push(from);
        while (open_wege.length > 0) {
            let w = [];
            const x = open_wege.pop();
            if (typeof (x) == "number") w.push(x);
            else w = x;
            const lastp = w[w.length - 1];
            const targets = connections[lastp];
            if (targets.length > 0) {
                for (let t = 0; t < targets.length; t++) {
                    const wn = [];
                    for (i = 0; i < w.length; i++) wn.push(w[i]);
                    wn.push(targets[t]);
                    if (targets[t] === to) {
                        closed_wege.push(wn);
                    } else {
                        open_wege.push(wn);
                    }
                }
            }
        }

        if (closed_wege.length > 0) return closed_wege[0];
        else return []
    }

    connected_nodes_on_node(node_id) {
        const ret = [];
        const r = node_id;
        for (let c = 0; c < this.connection_link_ids.cols; c++) {
            if (this.connection_link_ids.get(r, c) >= 0) ret.push(c);
        }
        return ret;
    }

    calc_multi_path() {
        let i;
        const fingerprints = [];
        for (i = 0; i < this.clicked_nodes.length; i++) fingerprints.push(this.nodes.getWithId(this.clicked_nodes[i]).getFingerprint());
        // get fommon node (and between fingerprints)

        let cf = fingerprints[0];
        for (i = 1; i < fingerprints.length; i++) {
            cf = Fingerprint.common_fingerprint(cf, fingerprints[i]);
        }

        const common_node = this.nodes.getWithFingerprint(cf);
        const nid = common_node.getId();

        // get all paths for selected nodes
        const path = [];
        for (i = 0; i < this.clicked_nodes.length; i++) {
            const w = this.get_from_to(nid, this.clicked_nodes[i]);
            if (w.length > 0) path.push(w);
        }

        for (i in path) {
            this.make_nodes_aktiv(path[i]);
        }

        // activate links
        const l = [];
        for (const p in path) {
            for (i = 1; i < path[p].length; i++) {
                l.push(this.link_between_nodes(path[p][i - 1], path[p][i]));
            }
        }
        this.make_links_aktiv(l);

        return common_node;
    }

    calc_single_path() {
        let cnr;
        let d;
        let n_id;
        const level = 25;

        const aktiv_node = this.clicked_nodes[0];
        let dist = [];
        let connected_n = [];
        const connected_l = [];
        let stack = [];


        // search row for connected nodes (target) and get link-nr	this.connection_link_ids has link_nr
        stack.push(aktiv_node);
        dist.push(0);
        while (stack.length > 0) {
            n_id = stack.pop();
            d = dist.pop();
            if (!connected_n.includes(n_id)) {
                connected_n.push(n_id);
                if (d < level) {	// more nodes possible
                    for (let c = 0; c < this.connection_link_ids.cols; c++) {
                        cnr = this.connection_link_ids.get(n_id, c);
                        if (cnr >= 0) {	// is connected
                            connected_l.push(cnr);
                            stack.push(c);
                            dist.push(d + 1);	// also search connection from this node
                        }
                    }
                }
            }
        }

        // search column for connected nodes (source) and get link-nr
        dist = [];
        stack = [];
        connected_n = connected_n.filter(function (value, index, arr) {
            return value !== aktiv_node;
        });	// remove aktiv node (pushed later)
        stack.push(aktiv_node);
        dist.push(0);
        while (stack.length > 0) {
            n_id = stack.pop();
            d = dist.pop();
            if (!connected_n.includes(n_id)) {
                connected_n.push(n_id);
                if (d < level) {
                    for (let r = 0; r < this.connection_link_ids.rows; r++) {
                        cnr = this.connection_link_ids.get(r, n_id);
                        if (cnr >= 0) {
                            connected_l.push(cnr)
                            stack.push(r);
                            dist.push(d + 1);	// new connected node
                        }
                    }
                }
            }
        }

        this.make_links_aktiv(connected_l)
        this.make_nodes_aktiv(connected_n)
    }

    make_nodes_aktiv(node_ids) {
        for (let i = 0; i < node_ids.length; i++) {
            this.nodes.getWithId(node_ids[i]).setActive(true);
        }
    }

    make_links_aktiv(link_ids) {
        for (let i = 0; i < link_ids.length; i++) {
            this.links_data[link_ids[i]].active = true;
        }
    }

    on_node_click(d) {
        this.deselect_links();
        this.deselect_nodes();
        if (d3.event.target.tagName !== "svg") {
            const pos = this.clicked_nodes.indexOf(d.id);
            if (pos < 0) {
                this.clicked_nodes.push(d.id);		// before clicked -> add
                this.fingerprints_view.addNode(d);
            } else {
                this.clicked_nodes.splice(pos, 1);	// was clicked -> remove
                this.fingerprints_view.removeNode(d);
            }
            console.log("clicked=" + this.clicked_nodes);
            if (this.clicked_nodes.length > 1) {
                var common_node = this.calc_multi_path();	// multible node -> shortest path
                this.fingerprints_view.setCommonNode(common_node);
            } else if (this.clicked_nodes.length === 1) {
                this.calc_single_path();// single node -> full path
                this.fingerprints_view.removeCommonNode();
            } else {	// last node per click deactivated
                this.fingerprints_view.removeAll();
            }
            if (this.clicked_nodes.length === 1) this.infobar.showinfoText("info", "click second point for common fingerprint or next to a point to clear")
        } else {	// clear all
            this.clicked_nodes = [];
            this.fingerprints_view.removeAll();
        }
        this.update_tree();
    }

    filter_by_fingerprint(fingerprint) {

        if (Fingerprint.is_zero(fingerprint)) {
            this.clear_all_filter();
        } else {
            for (const i in this.nodes_data) {
                this.nodes_data[i].filtered = Fingerprint.a_has_all_b(this.nodes_data[i].getFingerprint(), fingerprint);
            }
        }


        this.update_tree()
    }

    filter_by_frame(frame) {

        this.clear_all_filter();
        for (let i = 0; i < this.nodes_data.length; i++) {
            this.nodes_data[i].filtered = this.nodes_data[i].has_frame(frame);
        }

        this.update_tree()
    }

    clear_all_filter() {
        for (const i in this.nodes_data) {
            this.nodes_data[i].filtered = false;
        }
    }

    set_filtert_node(node_id) {
        this.clear_all_filter();
        const n = this.nodes_data.filter(function (d) {
            return d.id.toString() === node_id;
        })[0];
        n.filtered = true;
        this.update_tree()
    }

    set_filtert_path(frame, middle_step) {
        this.deselect_links();
        this.deselect_nodes();
        const l = this.clicked_nodes.length;
        const n = this.nodes_data.filter(function (d) {
            return d.node_frames.includes(frame);
        })[0];
        const node_id = n.getId();

        if (middle_step === false) {
            this.clicked_nodes.push(node_id);
            this.set_filtert_node(this.clicked_nodes[0]);
            if (l > 0) {
                this.set_filtert_node(this.clicked_nodes[1]);
                this.calc_multi_path();
            }
        } else {
            this.clicked_nodes = [];
            this.clicked_nodes.push(node_id);
            this.set_filtert_node(this.clicked_nodes[0]);
        }

        this.update_tree()
    }
}

//////////////////////////////
////////// tooltip ///////////
//////////////////////////////


TOOLTIPP_XOFFSET = 0;
TOOLTIPP_YOFFSET = -100;
TOOLTIPP_SETTINGS_WIDTH = 100;
TOOLTIPP_SETTINGS_HEIGHT = 100;


class NodeTooltip extends Fingerprint {

    constructor(feature_vector) {
        super(feature_vector);
        this.tooltip = d3.tip().attr("class", "tooltip_fingerprint").offset([TOOLTIPP_YOFFSET, TOOLTIPP_XOFFSET]).html("<div id='tipDiv'></div>");
    }

    get_tooltip() {
        return this.tooltip;
    }

    split_on_off(data) {
        const ret = [];
        for (let i in data) {
            ret.push({color: data[i].color, count: data[i].on, opacity: 1});
            ret.push({color: data[i].color, count: data[i].off, opacity: 0.2});
        }
        return ret;
    }

    draw(d) {
        this.tooltip.show();

        const tooltip_svg = d3.select("#tipDiv").append("svg")
            .attr("width", TOOLTIPP_SETTINGS_WIDTH)
            .attr("height", TOOLTIPP_SETTINGS_HEIGHT);


        let data = d.getFingerprintType().getFingerprintToColor(true);

        data = this.split_on_off(data);

        const arcs = d3.pie().sort(null).value(function (d) {
            return d.count;
        })(data);

        const arc = d3.arc(0, 0, 0, 0, 0, undefined)
            .outerRadius(TOOLTIPP_SETTINGS_HEIGHT / 2)
            .innerRadius(TOOLTIPP_SETTINGS_HEIGHT / 20)
            .padAngle(0.01)
            .cornerRadius(8);

        const pieG = tooltip_svg.selectAll("g")
            .data([data])
            .enter()
            .append("g")
            .attr("transform", "translate(" + [TOOLTIPP_SETTINGS_WIDTH / 2, TOOLTIPP_SETTINGS_HEIGHT / 2] + ")");

        const block = pieG.selectAll(".arc")
            .data(arcs);

        const newBlock = block.enter().append("g").classed("arc", true)
            .attr("stroke", "gray");

        newBlock.append("path")
            .attr("d", arc)
            .attr("id", function (d, i) {
                return "arc-" + i
            })
            .attr("fill", function (d) {
                return d.data.color;
            })
            .style("stroke-opacity", function (d) {
                return d.data.opacity;
            })
            .attr("fill-opacity", function (d) {
                return d.data.opacity;
            });

        d3.select(".tooltip_fingerprint").transition().delay(1).duration(500).style("opacity", 1);
    }

    move() {
        this.tooltip.style("left", (d3.event.pageX + TOOLTIPP_XOFFSET - 10) + "px")    // zittern -> 10: margin?
            .style("top", (d3.event.pageY + TOOLTIPP_YOFFSET - 10) + "px");
    }

    hide() {
        this.tooltip.hide();
    }
}